import logging
import os
import warnings
from functools import lru_cache as memoized
from typing import Tuple, Union
from urllib.parse import parse_qs, urlparse

import requests
import requests.auth
from urllib3.exceptions import InsecureRequestWarning

from .client import DEFAULT_TIMEOUT, GrafanaClient
from .elements import (
    Admin,
    Alerting,
    AlertingProvisioning,
    Annotations,
    Dashboard,
    DashboardVersions,
    Datasource,
    Folder,
    Health,
    Notifications,
    Organization,
    Organizations,
    Plugin,
    Rbac,
    Search,
    Snapshots,
    Teams,
    User,
    Users,
)
from .util import as_bool

logger = logging.getLogger(__name__)


class GrafanaApi:
    def __init__(
        self,
        auth=None,
        host="localhost",
        port=None,
        url_path_prefix="",
        protocol="http",
        verify=True,
        timeout=DEFAULT_TIMEOUT,
        user_agent: str = None,
    ):
        self.client = GrafanaClient(
            auth,
            host=host,
            port=port,
            url_path_prefix=url_path_prefix,
            protocol=protocol,
            verify=verify,
            timeout=timeout,
            user_agent=user_agent,
        )
        self.url = None
        self.admin = Admin(self.client)
        self.alerting = Alerting(self.client)
        self.alertingprovisioning = AlertingProvisioning(self.client)
        self.dashboard = Dashboard(self.client, self)
        self.dashboard_versions = DashboardVersions(self.client)
        self.datasource = Datasource(self.client, self)
        self.folder = Folder(self.client)
        self.health = Health(self.client)
        self.organization = Organization(self.client)
        self.organizations = Organizations(self.client, self)
        self.search = Search(self.client)
        self.user = User(self.client)
        self.users = Users(self.client)
        self.rbac = Rbac(self.client)
        self.teams = Teams(self.client)
        self.annotations = Annotations(self.client)
        self.snapshots = Snapshots(self.client)
        self.notifications = Notifications(self.client)
        self.plugin = Plugin(self.client)

    def connect(self):
        try:
            grafana_info = self.health.check()
        except requests.exceptions.ConnectionError as ex:
            logger.critical(f"Unable to connect to Grafana at {self.url or self.client.url_host}: {ex}")
            raise
        logger.info(f"Connected to Grafana at {self.url}: {grafana_info}")
        return grafana_info

    @property
    @memoized(maxsize=1)
    def version(self):
        grafana_info = self.health.check()
        version = grafana_info["version"]
        logger.info(f"Inquired Grafana version: {version}")
        return version

    @classmethod
    def from_url(
        cls,
        url: str = None,
        credential: Union[str, Tuple[str, str], requests.auth.AuthBase] = None,
        timeout: Union[float, Tuple[float, float]] = DEFAULT_TIMEOUT,
    ):
        """
        Factory method to create a `GrafanaApi` instance from a URL.

        Accepts an optional credential, which is either an authentication
        token, or a tuple of (username, password).
        """

        # Sanity checks and defaults.
        if url is None:
            url = "http://admin:admin@localhost:3000"

        if credential is not None and not isinstance(credential, (str, Tuple, requests.auth.AuthBase)):
            raise TypeError(f"Argument 'credential' has wrong type: {type(credential)}")

        original_url = url
        url = urlparse(url)

        # Use username and password from URL.
        if credential is None and url.username:
            credential = (url.username, url.password)

        # Optionally turn off SSL verification.
        verify = as_bool(parse_qs(url.query).get("verify", [True])[0])
        if verify is False:
            warnings.filterwarnings("ignore", category=InsecureRequestWarning)

        grafana = cls(
            credential,
            protocol=url.scheme,
            host=url.hostname,
            port=url.port,
            url_path_prefix=url.path.lstrip("/"),
            verify=verify,
            timeout=timeout,
        )
        grafana.url = original_url

        return grafana

    @classmethod
    def from_env(cls, timeout: Union[float, Tuple[float, float]] = None):
        """
        Factory method to create a `GrafanaApi` instance from environment variables.
        """
        if timeout is None:
            if "GRAFANA_TIMEOUT" in os.environ:
                try:
                    timeout = float(os.environ["GRAFANA_TIMEOUT"])
                except Exception as ex:
                    raise ValueError(
                        f"Unable to parse invalid `float` value from " f"`GRAFANA_TIMEOUT` environment variable: {ex}"
                    )
        if timeout is None:
            timeout = DEFAULT_TIMEOUT
        return cls.from_url(
            url=os.environ.get("GRAFANA_URL"), credential=os.environ.get("GRAFANA_TOKEN"), timeout=timeout
        )



from json import JSONDecodeError

import requests
import requests.auth

DEFAULT_TIMEOUT: float = 5.0


class GrafanaException(Exception):
    def __init__(self, status_code, response, message):
        self.status_code = status_code
        self.response = response
        self.message = message
        # Backwards compatible with implementations that rely on just the message.
        super(GrafanaException, self).__init__(message)


class GrafanaServerError(GrafanaException):
    """
    5xx
    """

    pass


class GrafanaClientError(GrafanaException):
    """
    Invalid input (4xx errors)
    """

    pass


class GrafanaBadInputError(GrafanaClientError):
    """
    400
    """

    def __init__(self, response):
        super(GrafanaBadInputError, self).__init__(400, response, "Bad Input: `{0}`".format(response))


class GrafanaUnauthorizedError(GrafanaClientError):
    """
    401
    """

    def __init__(self, response):
        super(GrafanaUnauthorizedError, self).__init__(401, response, "Unauthorized")


class TokenAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, request):
        request.headers.update({"Authorization": "Bearer {0}".format(self.token)})
        return request


class HeaderAuth(requests.auth.AuthBase):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __call__(self, request):
        request.headers.update({self.name: self.value})
        return request


class GrafanaClient:
    def __init__(
        self,
        auth,
        host="localhost",
        port=None,
        url_path_prefix="",
        protocol="http",
        verify=True,
        timeout=DEFAULT_TIMEOUT,
        user_agent: str = None,
    ):
        self.auth = auth
        self.verify = verify
        self.timeout = timeout
        self.url_host = host
        self.url_port = port
        self.url_path_prefix = url_path_prefix
        self.url_protocol = protocol

        def construct_api_url():
            params = {
                "protocol": self.url_protocol,
                "host": self.url_host,
                "url_path_prefix": self.url_path_prefix,
            }

            if self.url_port is None:
                url_pattern = "{protocol}://{host}/{url_path_prefix}api"
            else:
                params["port"] = self.url_port
                url_pattern = "{protocol}://{host}:{port}/{url_path_prefix}api"

            return url_pattern.format(**params)

        self.url = construct_api_url()

        from grafana_client import __appname__, __version__

        self.user_agent = user_agent or f"{__appname__}/{__version__}"

        self.s = requests.Session()
        self.s.headers["User-Agent"] = self.user_agent
        if self.auth is not None:
            if isinstance(self.auth, requests.auth.AuthBase):
                pass
            elif isinstance(self.auth, tuple):
                self.auth = requests.auth.HTTPBasicAuth(*self.auth)
            else:
                self.auth = TokenAuth(self.auth)

    def __getattr__(self, item):
        def __request_runner(url, json=None, data=None, headers=None, accept_empty_json=False):
            __url = "%s%s" % (self.url, url)
            # Sanity checks.
            if json is not None and not isinstance(json, (dict, list)):
                raise TypeError(
                    f"JSON request payload has invalid shape. "
                    f"Accepted are dictionaries and lists. "
                    f"The type is: {type(json)}"
                )
            runner = getattr(self.s, item.lower())
            r = runner(
                __url,
                json=json,
                data=data,
                headers=headers,
                auth=self.auth,
                verify=self.verify,
                timeout=self.timeout,
            )
            if r.status_code >= 400:
                try:
                    response = r.json()
                except ValueError:
                    response = r.text
                message = response["message"] if isinstance(response, dict) and "message" in response else r.text

                if 500 <= r.status_code < 600:
                    raise GrafanaServerError(
                        r.status_code,
                        response,
                        "Server Error {0}: {1}".format(r.status_code, message),
                    )
                elif r.status_code == 400:
                    raise GrafanaBadInputError(response)
                elif r.status_code == 401:
                    raise GrafanaUnauthorizedError(response)
                elif 400 <= r.status_code < 500:
                    raise GrafanaClientError(
                        r.status_code,
                        response,
                        "Client Error {0}: {1}".format(r.status_code, message),
                    )

            # `204 No Content` responses have an empty response body,
            # so it doesn't decode well from JSON.
            if r.status_code == 204:
                return None

            # The "Tempo" data source responds with text/plain.
            content_type = r.headers.get("Content-Type", "")
            if content_type.startswith("text/"):
                return r.text
            try:
                return r.json()
            except JSONDecodeError:
                if accept_empty_json and r.text == "":
                    return ""
                else:
                    raise

        return __request_runner



import unittest

import requests_mock

from grafana_client import GrafanaApi

ALERTRULE = {
    "name": "alert-rule-test",
    "id": 2,
    "uid": "bUUGqLiVk",
    "orgID": 1,
    "folderUID": "4",
    "ruleGroup": "Any",
    "title": "ApiAlerts",
    "condition": "B",
    "data": [],
    "updated": "2022-08-17T13:00:21Z",
    "noDataState": "NoData",
    "execErrState": "Alerting",
    "for": 300000000000,
    "labels": {"alertGroup": "any", "region": "eu-west-1"},
}


class AlertingTestCase(unittest.TestCase):
    def setUp(self):
        self.grafana = GrafanaApi(("admin", "admin"), host="localhost", url_path_prefix="", protocol="http")

    @requests_mock.Mocker()
    def test_get_alertrule(self, m):
        m.get("http://localhost/api/ruler/grafana/api/v1/rules/alert-folder/alert-rule-test", json=ALERTRULE)
        response = self.grafana.alerting.get_alertrule("alert-folder", "alert-rule-test")
        self.assertEqual(response["uid"], "bUUGqLiVk")
        self.assertEqual(response["name"], "alert-rule-test")

    @requests_mock.Mocker()
    def test_delete_alertrule(self, m):
        m.delete(
            "http://localhost/api/ruler/grafana/api/v1/rules/alert-folder/alert-rule-test", json={"uid": "bUUGqLiVk"}
        )
        response = self.grafana.alerting.delete_alertrule("alert-folder", "alert-rule-test")
        self.assertEqual(response["uid"], "bUUGqLiVk")

    @requests_mock.Mocker()
    def test_create_alertrule(self, m):
        m.post(
            "http://localhost/api/ruler/grafana/api/v1/rules/alert-folder",
            json=ALERTRULE,
        )

        response = self.grafana.alerting.create_alertrule("alert-folder", ALERTRULE)
        self.assertEqual(response["uid"], "bUUGqLiVk")
        self.assertEqual(response["name"], "alert-rule-test")

    @requests_mock.Mocker()
    def test_update_alertrule(self, m):
        m.post(
            "http://localhost/api/ruler/grafana/api/v1/rules/alert-folder",
            json=ALERTRULE,
        )

        response = self.grafana.alerting.update_alertrule("alert-folder", alertrule=ALERTRULE)
        self.assertEqual(response["uid"], "bUUGqLiVk")



import unittest
from test.elements.test_datasource_fixtures import (
    DATAFRAME_RESPONSE_EMPTY,
    DATAFRAME_RESPONSE_HEALTH_PROMETHEUS,
    DATAFRAME_RESPONSE_HEALTH_SELECT1,
    DATAFRAME_RESPONSE_INVALID,
    ELASTICSEARCH_DATASOURCE,
    GRAPHITE_DATASOURCE,
    INFLUXDB1_DATASOURCE,
    JAEGER_DATASOURCE,
    LOKI_DATASOURCE,
    MSSQL_DATASOURCE,
    MYSQL_DATASOURCE,
    OPENTSDB_DATASOURCE,
    POSTGRES_DATASOURCE,
    PROMETHEUS_DATASOURCE,
    SIMPLE_JSON_DATASOURCE,
    SIMPOD_JSON_DATASOURCE,
    SUNANDMOON_DATASOURCE,
    SUNANDMOON_DATASOURCE_INCOMPLETE,
    TEMPO_DATASOURCE,
    TESTDATA_DATASOURCE,
    ZIPKIN_DATASOURCE,
)

import requests_mock
from parameterized import parameterized

from grafana_client import GrafanaApi
from grafana_client.client import GrafanaClientError, GrafanaServerError
from grafana_client.model import DatasourceHealthResponse, DatasourceIdentifier


class DatasourceHealthCheckTestCase(unittest.TestCase):
    def setUp(self):
        self.grafana = GrafanaApi(("admin", "admin"), host="localhost", url_path_prefix="", protocol="http")

    @requests_mock.Mocker()
    def test_health_check_access_type_failure(self, m):
        m.get(
            "http://localhost/api/datasources/uid/h8KkCLt7z",
            json=PROMETHEUS_DATASOURCE,
        )

        datasource = PROMETHEUS_DATASOURCE.copy()
        datasource["access"] = "__UNKNOWN__"
        self.assertRaises(NotImplementedError, lambda: self.grafana.datasource.health_check(datasource))

    @requests_mock.Mocker()
    def test_health_check_elasticsearch_success(self, m):
        m.get(
            "http://localhost/api/datasources/uid/34inf2sdc",
            json=ELASTICSEARCH_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"bazqux": {"mappings": {"properties": "something"}}},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="34inf2sdc"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="34inf2sdc",
                type="elasticsearch",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_elasticsearch_empty_response_failure(self, m):
        m.get(
            "http://localhost/api/datasources/uid/34inf2sdc",
            json=ELASTICSEARCH_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="34inf2sdc"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="34inf2sdc",
                type="elasticsearch",
                success=False,
                status="ERROR",
                message="No response for database 'bazqux'",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_elasticsearch_incomplete_response_failure(self, m):
        m.get(
            "http://localhost/api/datasources/uid/34inf2sdc",
            json=ELASTICSEARCH_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"bazqux": {}},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="34inf2sdc"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="34inf2sdc",
                type="elasticsearch",
                success=False,
                status="ERROR",
                message="Invalid response. KeyError: 'mappings'",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_elasticsearch_error_response_failure(self, m):
        m.get(
            "http://localhost/api/datasources/uid/34inf2sdc",
            json=ELASTICSEARCH_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"error": "This failed!", "status": 400},
            status_code=400,
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="34inf2sdc"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="34inf2sdc",
                type="elasticsearch",
                success=False,
                status="ERROR",
                message="This failed!",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_elasticsearch_error_response_with_root_cause_failure(self, m):
        m.get(
            "http://localhost/api/datasources/uid/34inf2sdc",
            json=ELASTICSEARCH_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"error": {"root_cause": [{"type": "foo", "reason": "bar"}]}, "status": 400},
            status_code=400,
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="34inf2sdc"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="34inf2sdc",
                type="elasticsearch",
                success=False,
                status="ERROR",
                message="Status: 400. Type: foo. Reason: bar",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_graphite_success(self, m):
        m.get(
            "http://localhost/api/datasources/uid/wr47rz34e",
            json=GRAPHITE_DATASOURCE,
        )
        m.post(
            "http://localhost/api/datasources/proxy/48/render",
            json=[{"target": "", "tags": {}, "datapoints": []}],
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="wr47rz34e"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="wr47rz34e",
                type="graphite",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_graphite_empty_response_failure(self, m):
        m.get(
            "http://localhost/api/datasources/uid/wr47rz34e",
            json=GRAPHITE_DATASOURCE,
        )
        m.post(
            "http://localhost/api/datasources/proxy/48/render",
            json=[],
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="wr47rz34e"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="wr47rz34e",
                type="graphite",
                success=False,
                status="ERROR",
                message="Invalid response. IndexError: list index out of range",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_influxdb1(self, m):
        m.get(
            "http://localhost/api/datasources/uid/9ac78sdc2",
            json=INFLUXDB1_DATASOURCE,
        )
        m.post(
            "http://localhost/api/datasources/proxy/43/query",
            json={
                "results": [
                    {
                        "statement_id": 0,
                        "series": [
                            {
                                "columns": ["name", "duration", "shardGroupDuration", "replicaN", "default"],
                                "values": [["monitor", "168h0m0s", "24h0m0s", 1, True]],
                            }
                        ],
                    }
                ]
            },
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="9ac78sdc2"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="9ac78sdc2",
                type="influxdb",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_jaeger_success(self, m):
        m.get(
            "http://localhost/api/datasources/uid/DbtFe237k",
            json=JAEGER_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"data": ["jaeger-query"], "total": 1, "limit": 0, "offset": 0, "errors": None},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="DbtFe237k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="DbtFe237k",
                type="jaeger",
                success=True,
                status="OK",
                message="['jaeger-query']",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_jaeger_error_response_failure(self, m):
        m.get(
            "http://localhost/api/datasources/uid/DbtFe237k",
            json=JAEGER_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={
                "data": ["jaeger-query"],
                "total": 1,
                "limit": 0,
                "offset": 0,
                "errors": [{"code": 418, "message": "foobar"}],
            },
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="DbtFe237k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="DbtFe237k",
                type="jaeger",
                success=False,
                status="ERROR",
                message="[{'code': 418, 'message': 'foobar'}]",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_loki_success_grafana7(self, m):
        # Mock the version inquiry request, because `smartquery` needs
        # it, as Loki responses differ between versions.
        m.get(
            "http://localhost/api/health",
            json={"commit": "unknown", "database": "ok", "version": "7.0.1"},
        )
        m.get(
            "http://localhost/api/datasources/uid/vCyglaq7z",
            json=LOKI_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"status": "success", "data": ["__name__"]},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="vCyglaq7z"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="vCyglaq7z",
                type="loki",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_loki_success_grafana9(self, m):
        # Mock the version inquiry request, because `smartquery` needs
        # it, as Loki responses differ between versions.
        m.get(
            "http://localhost/api/health",
            json={"commit": "14e988bd22", "database": "ok", "version": "9.0.1"},
        )
        m.get(
            "http://localhost/api/datasources/uid/vCyglaq7z",
            json=LOKI_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"results": {"test": True}},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="vCyglaq7z"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="vCyglaq7z",
                type="loki",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_loki_error_response_failure_grafana7(self, m):
        # Mock the version inquiry request, because `smartquery` needs
        # it, as Loki responses differ between versions.
        m.get(
            "http://localhost/api/health",
            json={"commit": "unknown", "database": "ok", "version": "7.0.1"},
        )
        m.get(
            "http://localhost/api/datasources/uid/vCyglaq7z",
            json=LOKI_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"message": "Failed to call resource", "traceID": "00000000000000000000000000000000"},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="vCyglaq7z"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="vCyglaq7z",
                type="loki",
                success=False,
                status="ERROR",
                message="Failed to call resource",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_loki_error_response_failure_grafana9(self, m):
        # Mock the version inquiry request, because `smartquery` needs
        # it, as Loki responses differ between versions.
        m.get(
            "http://localhost/api/health",
            json={"commit": "14e988bd22", "database": "ok", "version": "9.0.1"},
        )
        m.get(
            "http://localhost/api/datasources/uid/vCyglaq7z",
            json=LOKI_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"message": "Failed to call resource", "traceID": "00000000000000000000000000000000"},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="vCyglaq7z"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="vCyglaq7z",
                type="loki",
                success=False,
                status="ERROR",
                message="Failed to call resource",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_mssql(self, m):
        m.get(
            "http://localhost/api/datasources/uid/0pueH83nz",
            json=MSSQL_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json=DATAFRAME_RESPONSE_HEALTH_SELECT1,
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="0pueH83nz"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="0pueH83nz",
                type="mssql",
                success=True,
                status="OK",
                message="SELECT 1",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_mysql(self, m):
        m.get(
            "http://localhost/api/datasources/uid/7CpzLp37z",
            json=MYSQL_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json=DATAFRAME_RESPONSE_HEALTH_SELECT1,
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="7CpzLp37z"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="7CpzLp37z",
                type="mysql",
                success=True,
                status="OK",
                message="SELECT 1",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_opentsdb(self, m):
        m.get(
            "http://localhost/api/datasources/uid/hkuk5h3nk",
            json=OPENTSDB_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json="",
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="hkuk5h3nk"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="hkuk5h3nk",
                type="opentsdb",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_postgres(self, m):
        m.get(
            "http://localhost/api/datasources/uid/v2KYBt37k",
            json=POSTGRES_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json=DATAFRAME_RESPONSE_HEALTH_SELECT1,
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="v2KYBt37k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="v2KYBt37k",
                type="postgres",
                success=True,
                status="OK",
                message="SELECT 1",
                duration=None,
                response=None,
            ),
        )

    @parameterized.expand(["7.0.1", "9.0.1"])
    def test_health_check_prometheus_healthy_success(self, grafana_version):
        m = requests_mock.Mocker()
        # Mock the version inquiry request, because `smartquery` needs
        # it, as Prometheus responses differ between versions.
        m.get(
            "http://localhost/api/health",
            json={"commit": "unknown", "database": "ok", "version": grafana_version},
        )
        m.get(
            "http://localhost/api/datasources/uid/h8KkCLt7z",
            json=PROMETHEUS_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json=DATAFRAME_RESPONSE_HEALTH_PROMETHEUS,
        )
        with m:
            response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="h8KkCLt7z"))
            response.duration = None
            response.response = None
            self.assertEqual(
                response,
                DatasourceHealthResponse(
                    uid="h8KkCLt7z",
                    type="prometheus",
                    success=True,
                    status="OK",
                    message="Expr: 1+1\nStep: 15s",
                    duration=None,
                    response=None,
                ),
            )

    @parameterized.expand(["7.0.1", "9.0.1"])
    def test_health_check_prometheus_empty_dataframe_success(self, grafana_version):
        m = requests_mock.Mocker()
        # Mock the version inquiry request, because `smartquery` needs
        # it, as Prometheus responses differ between versions.
        m.get(
            "http://localhost/api/health",
            json={"commit": "unknown", "database": "ok", "version": grafana_version},
        )
        m.get(
            "http://localhost/api/datasources/uid/h8KkCLt7z",
            json=PROMETHEUS_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json=DATAFRAME_RESPONSE_EMPTY,
        )
        with m:
            response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="h8KkCLt7z"))
            response.duration = None
            response.response = None
            self.assertEqual(
                response,
                DatasourceHealthResponse(
                    uid="h8KkCLt7z",
                    type="prometheus",
                    success=True,
                    status="OK",
                    message="Success",
                    duration=None,
                    response=None,
                ),
            )

    @parameterized.expand(["7.0.1", "9.0.1"])
    def test_health_check_prometheus_invalid_dataframe_failure(self, grafana_version):
        m = requests_mock.Mocker()
        # Mock the version inquiry request, because `smartquery` needs
        # it, as Prometheus responses differ between versions.
        m.get(
            "http://localhost/api/health",
            json={"commit": "unknown", "database": "ok", "version": grafana_version},
        )
        m.get(
            "http://localhost/api/datasources/uid/h8KkCLt7z",
            json=PROMETHEUS_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json=DATAFRAME_RESPONSE_INVALID,
        )
        with m:
            response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="h8KkCLt7z"))
            response.duration = None
            response.response = None
            self.assertEqual(
                response,
                DatasourceHealthResponse(
                    uid="h8KkCLt7z",
                    type="prometheus",
                    success=False,
                    status="ERROR",
                    message="FATAL: Unable to decode result from dictionary-type response. "
                    "TypeError: DataFrame response detected, but 'frames' is not a list",
                    duration=None,
                    response=None,
                ),
            )

    @requests_mock.Mocker()
    def test_health_check_simplejson(self, m):
        m.get(
            "http://localhost/api/datasources/uid/rw783ds3e",
            json=SIMPLE_JSON_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"results": {"test": {"error": "Resource not found: /metadata"}}},
            status_code=404,
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="rw783ds3e"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="rw783ds3e",
                type="grafana-simple-json-datasource",
                success=False,
                status="ERROR",
                message="Resource not found: /metadata",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_simpod_json(self, m):
        m.get(
            "http://localhost/api/datasources/uid/oie238af3",
            json=SIMPOD_JSON_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"results": [{"statement_id": "ID", "series": []}]},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="oie238af3"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="oie238af3",
                type="simpod-json-datasource",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_sunandmoon_success(self, m):
        m.get(
            "http://localhost/api/datasources/uid/239fasva4",
            json=SUNANDMOON_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json=SUNANDMOON_DATASOURCE,
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="239fasva4"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="239fasva4",
                type="fetzerch-sunandmoon-datasource",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_sunandmoon_incomplete_failure(self, m):
        m.get(
            "http://localhost/api/datasources/uid/239fasva4",
            json=SUNANDMOON_DATASOURCE_INCOMPLETE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json=SUNANDMOON_DATASOURCE_INCOMPLETE,
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="239fasva4"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="239fasva4",
                type="fetzerch-sunandmoon-datasource",
                success=False,
                status="ERROR",
                message="Invalid response. KeyError: 'latitude'",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_tempo_success(self, m):
        m.get(
            "http://localhost/api/datasources/uid/aTk86s3nk",
            json=TEMPO_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            headers={"Content-Type": "text/plain"},
            text="echo",
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="aTk86s3nk"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="aTk86s3nk",
                type="tempo",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_tempo_error_response_failure(self, m):
        m.get(
            "http://localhost/api/datasources/uid/aTk86s3nk",
            json=TEMPO_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            status_code=502,
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="aTk86s3nk"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="aTk86s3nk",
                type="tempo",
                success=False,
                status="ERROR",
                message="Server Error 502: ",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_testdata(self, m):
        m.get(
            "http://localhost/api/datasources/uid/439fngqr2",
            json=TESTDATA_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"id": "test", "uid": "test"},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="439fngqr2"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="439fngqr2",
                type="testdata",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_zipkin_success(self, m):
        m.get(
            "http://localhost/api/datasources/uid/3sXIv8q7k",
            json=ZIPKIN_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json=[],
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="3sXIv8q7k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="3sXIv8q7k",
                type="zipkin",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_zdict_valid_response_success(self, m):
        """
        This is not a test for a real `postgres` data source.
        It only exercises a specific response shape.
        Here: The new DataFrame dictionary-type response.
        """
        m.get(
            "http://localhost/api/datasources/uid/v2KYBt37k",
            json=POSTGRES_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"results": {"test": {"refId": "foobar", "meta": {"executedQueryString": "ALIVE?"}}}},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="v2KYBt37k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="v2KYBt37k",
                type="postgres",
                success=True,
                status="OK",
                message="ALIVE?",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_zdict_incomplete_response_success(self, m):
        """
        This is not a test for a real `postgres` data source.
        It only exercises a specific response shape.
        Here: The new DataFrame dictionary-type response.
        """
        m.get(
            "http://localhost/api/datasources/uid/v2KYBt37k",
            json=POSTGRES_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"results": {"test": {"refId": "foobar"}}},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="v2KYBt37k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="v2KYBt37k",
                type="postgres",
                success=True,
                status="OK",
                message="Success",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_zdict_incomplete_response_failure(self, m):
        """
        This is not a test for a real `postgres` data source.
        It only exercises a specific response shape.
        Here: The new DataFrame dictionary-type response.
        """
        m.get(
            "http://localhost/api/datasources/uid/v2KYBt37k",
            json=POSTGRES_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"results": {"test": {}}},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="v2KYBt37k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="v2KYBt37k",
                type="postgres",
                success=False,
                status="ERROR",
                message="FATAL: Unable to decode result from dictionary-type response. "
                "TypeError: Invalid response format",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_zlist_error_response_failure(self, m):
        """
        This is not a test for a real `postgres` data source.
        It only exercises a specific response shape.
        Here: The previous list-type response.
        """
        m.get(
            "http://localhost/api/datasources/uid/v2KYBt37k",
            json=POSTGRES_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"results": [{"error": "This failed!"}]},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="v2KYBt37k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="v2KYBt37k",
                type="postgres",
                success=False,
                status="ERROR",
                message="This failed!",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_zlist_incomplete_response_failure(self, m):
        """
        This is not a test for a real `postgres` data source.
        It only exercises a specific response shape.
        Here: The previous list-type response.
        """
        m.get(
            "http://localhost/api/datasources/uid/v2KYBt37k",
            json=POSTGRES_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"results": [{}]},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="v2KYBt37k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="v2KYBt37k",
                type="postgres",
                success=False,
                status="ERROR",
                message="FATAL: Unable to decode result from list-type response. KeyError: 'statement_id'",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_zlist_empty_response_failure(self, m):
        """
        This is not a test for a real `postgres` data source.
        It only exercises a specific response shape.
        Here: A response with an empty "results" slot.
        """
        m.get(
            "http://localhost/api/datasources/uid/v2KYBt37k",
            json=POSTGRES_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"results": []},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="v2KYBt37k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="v2KYBt37k",
                type="postgres",
                success=False,
                status="ERROR",
                message="FATAL: Unable to decode result from list-type response. IndexError: list index out of range",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_zdummy_void_response_failure(self, m):
        """
        This is not a test for a real `postgres` data source.
        It only exercises a specific response shape.
        Here: An absolute empty (void) response.
        """
        m.get(
            "http://localhost/api/datasources/uid/v2KYBt37k",
            json=POSTGRES_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="v2KYBt37k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="v2KYBt37k",
                type="postgres",
                success=False,
                status="ERROR",
                message="Response lacks expected keys 'results' or 'data'",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_zzdummy_invalid_type_failure(self, m):
        """
        This is not a test for a real `postgres` data source.
        It only exercises a specific response shape.
        Here: A response with an invalid data type.
        """
        m.get(
            "http://localhost/api/datasources/uid/v2KYBt37k",
            json=POSTGRES_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"results": "WRONG!"},
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="v2KYBt37k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="v2KYBt37k",
                type="postgres",
                success=False,
                status="ERROR",
                message="FATAL: Unknown response type '<class 'str'>'. Expected: dictionary or list.",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_check_zzdummy_exception_without_response_failure(self, m):
        """
        This is not a test for a real `postgres` data source.
        It only exercises a specific response shape.
        Here: A response with an invalid data type.
        """
        m.get(
            "http://localhost/api/datasources/uid/v2KYBt37k",
            json=POSTGRES_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json=None,
            status_code=400,
        )
        response = self.grafana.datasource.health_check(DatasourceIdentifier(uid="v2KYBt37k"))
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="v2KYBt37k",
                type="postgres",
                success=False,
                status="ERROR",
                message="Bad Input: ``",
                duration=None,
                response=None,
            ),
        )


class DatasourceHealthInquiryTestCase(unittest.TestCase):
    def setUp(self):
        self.grafana = GrafanaApi(("admin", "admin"), host="localhost", url_path_prefix="", protocol="http")

    @requests_mock.Mocker()
    def test_health_inquiry_native_prometheus_success(self, m):
        m.get(
            "http://localhost/api/health",
            json={"commit": "14e988bd22", "database": "ok", "version": "9.0.1"},
        )
        m.get(
            "http://localhost/api/datasources/uid/39mf288en",
            json={"type": "unknown"},
        )
        m.get(
            "http://localhost/api/datasources/uid/39mf288en/health",
            json={"status": "OK", "message": "Excellent!"},
        )
        response = self.grafana.datasource.health_inquiry(datasource_uid="39mf288en")
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="39mf288en",
                type="unknown",
                success=True,
                status="OK",
                message="Excellent!",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_inquiry_native_prometheus_failure(self, m):
        m.get(
            "http://localhost/api/health",
            json={"commit": "14e988bd22", "database": "ok", "version": "9.0.1"},
        )
        m.get(
            "http://localhost/api/datasources/uid/39mf288en",
            json={"type": "unknown"},
        )
        m.get(
            "http://localhost/api/datasources/uid/39mf288en/health",
            json={"status": "ERROR", "message": "No way!"},
        )
        response = self.grafana.datasource.health_inquiry(datasource_uid="39mf288en")
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="39mf288en",
                type="unknown",
                success=False,
                status="ERROR",
                message="No way!",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_inquiry_native_unknown_error_400(self, m):
        m.get(
            "http://localhost/api/health",
            json={"commit": "14e988bd22", "database": "ok", "version": "9.0.1"},
        )
        m.get(
            "http://localhost/api/datasources/uid/39mf288en",
            json={"type": "unknown"},
        )
        m.get(
            "http://localhost/api/datasources/uid/39mf288en/health",
            json={"status": "ERROR", "message": "Something failed"},
            status_code=400,
        )
        response = self.grafana.datasource.health_inquiry(datasource_uid="39mf288en")
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="39mf288en",
                type="unknown",
                success=False,
                status="ERROR",
                message="Something failed",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_inquiry_native_unknown_error_418(self, m):
        m.get(
            "http://localhost/api/health",
            json={"commit": "14e988bd22", "database": "ok", "version": "9.0.1"},
        )
        m.get(
            "http://localhost/api/datasources/uid/39mf288en",
            json={"type": "unknown"},
        )
        m.get(
            "http://localhost/api/datasources/uid/39mf288en/health",
            json={"status": "ERROR", "message": "Something failed"},
            status_code=418,
        )
        with self.assertRaises(GrafanaClientError) as ctx:
            self.grafana.datasource.health_inquiry(datasource_uid="39mf288en")
        self.assertEqual(str(ctx.exception), "Client Error 418: Something failed")

    @requests_mock.Mocker()
    def test_health_inquiry_native_prometheus_error_404(self, m):
        m.get(
            "http://localhost/api/health",
            json={"commit": "14e988bd22", "database": "ok", "version": "9.0.1"},
        )
        m.get(
            "http://localhost/api/datasources/uid/h8KkCLt7z",
            json=PROMETHEUS_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"status": "ERROR", "message": "Something failed"},
            status_code=404,
        )
        m.get(
            "http://localhost/api/datasources/uid/h8KkCLt7z/health",
            json={"status": "ERROR", "message": "Something failed"},
            status_code=404,
        )
        response = self.grafana.datasource.health_inquiry(datasource_uid="h8KkCLt7z")
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="h8KkCLt7z",
                type="prometheus",
                success=False,
                status="ERROR",
                message="Unknown: Client Error 404: Something failed. "
                "Response: {'status': 'ERROR', 'message': 'Something failed'}",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_inquiry_native_prometheus_error_500(self, m):
        m.get(
            "http://localhost/api/health",
            json={"commit": "14e988bd22", "database": "ok", "version": "9.0.1"},
        )
        m.get(
            "http://localhost/api/datasources/uid/h8KkCLt7z",
            json=PROMETHEUS_DATASOURCE,
        )
        m.post(
            "http://localhost/api/ds/query",
            json={"status": "ERROR", "message": "Something failed"},
            status_code=500,
        )
        m.get(
            "http://localhost/api/datasources/uid/h8KkCLt7z/health",
            json={"status": "ERROR", "message": "Something failed", "code": "foobar"},
            status_code=500,
        )
        response = self.grafana.datasource.health_inquiry(datasource_uid="h8KkCLt7z")
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="h8KkCLt7z",
                type="prometheus",
                success=False,
                status="FATAL",
                message="[foobar] GrafanaServerError: Server Error 500: Something failed",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_inquiry_response_404(self, m):
        m.get(
            "http://localhost/api/datasources/uid/39mf288en",
            json={"status": "FATAL", "message": "Not found"},
            status_code=404,
        )
        response = self.grafana.datasource.health_inquiry(datasource_uid="39mf288en")
        response.duration = None
        response.response = None
        self.assertEqual(
            response,
            DatasourceHealthResponse(
                uid="39mf288en",
                type=None,
                success=False,
                status="ERROR",
                message="Not found",
                duration=None,
                response=None,
            ),
        )

    @requests_mock.Mocker()
    def test_health_inquiry_response_418(self, m):
        m.get(
            "http://localhost/api/datasources/uid/39mf288en",
            json={"status": "FATAL", "message": "Bad request"},
            status_code=418,
        )
        self.assertRaises(
            GrafanaClientError, lambda: self.grafana.datasource.health_inquiry(datasource_uid="39mf288en")
        )

    @requests_mock.Mocker()
    def test_health_inquiry_response_500(self, m):
        m.get(
            "http://localhost/api/datasources/uid/39mf288en",
            json={"status": "FATAL", "message": "Server failed"},
            status_code=500,
        )
        self.assertRaises(
            GrafanaServerError, lambda: self.grafana.datasource.health_inquiry(datasource_uid="39mf288en")
        )



"""
About
=====

Example program demonstrating how to work with both `grafana-client`_ and
`grafanalib`_. It reflects the example program `example.upload-dashboard.py`_
from `grafanalib`_ and will create and upload a dashboard.

Setup
=====
::

    pip install grafana-client grafanalib

Synopsis
========
::

    # Invoke example program.
    # By default, it will assume a Grafana instance accessible
    # via `http://admin:admin@localhost:3000`.
    python examples/grafanalib-upload-dashboard.py

    # Optionally target a remote Grafana instance, with authentication token.
    export GRAFANA_URL=https://grafana.example.org/
    export GRAFANA_TOKEN=eyJrIjoiWHg...dGJpZCI6MX0=

.. _example.upload-dashboard.py: https://github.com/weaveworks/grafanalib/blob/main/grafanalib/tests/examples/example.upload-dashboard.py
.. _grafana-client: https://github.com/panodata/grafana-client
.. _grafanalib: https://github.com/weaveworks/grafanalib
"""  # noqa:E501
import json
import logging
from typing import Dict

from grafanalib._gen import DashboardEncoder
from grafanalib.core import Dashboard

from grafana_client import GrafanaApi
from grafana_client.util import setup_logging

logger = logging.getLogger(__name__)


def mkdashboard(uid: str, title: str = None, message: str = None, overwrite: bool = False) -> Dict:
    """
    Create a dashboard create/update payload using grafanalib, suitable for
    submitting to the Grafana HTTP API.

    Note: This is solely for demonstration purposes, a real-world
    implementation would stuff more details into the dashboard beforehand.
    """
    dashboard = Dashboard(uid=uid, title=title)
    data = {
        "dashboard": encode_dashboard(dashboard),
        "overwrite": overwrite,
        "message": message,
    }
    return data


def encode_dashboard(entity) -> Dict:
    """
    Encode grafanalib `Dashboard` entity to dictionary.

    TODO: Optimize without going through JSON marshalling.
    """
    return json.loads(json.dumps(entity, sort_keys=True, cls=DashboardEncoder))


def run(grafana: GrafanaApi):
    """
    The main body of the example program.

    - Create a Grafana dashboard create/update payload.
    - Submit to Grafana HTTP API.
    - Display response.
    """

    # Create dashboard payload.
    dashboard_payload = mkdashboard(
        uid="abifsd",
        title="My awesome dashboard",
        message="Updated by grafanalib",
        overwrite=True,
    )

    # Create or update dashboard at Grafana HTTP API.
    response = grafana.dashboard.update_dashboard(dashboard_payload)

    # Display the outcome in JSON format.
    print(json.dumps(response, indent=2))


if __name__ == "__main__":
    """
    Boilerplate bootloader. Create a `GrafanaApi` instance and run example.
    """

    # Setup logging.
    setup_logging(level=logging.DEBUG)

    # Create a `GrafanaApi` instance, optionally configured with environment
    # variables `GRAFANA_URL` and `GRAFANA_TOKEN`.
    grafana_client = GrafanaApi.from_env()

    # Invoke example conversation.
    run(grafana_client)



import os

# Purge environment variables to prevent them leaking from the user environment
# into the test suite.
PURGE_ENV_VARS = ["GRAFANA_URL", "GRAFANA_TOKEN"]

for envvar in PURGE_ENV_VARS:
    if envvar in os.environ:
        del os.environ[envvar]



import unittest

import requests_mock

from grafana_client import GrafanaApi


class PluginTestCase(unittest.TestCase):
    def setUp(self):
        self.grafana = GrafanaApi(("admin", "admin"), host="localhost", url_path_prefix="", protocol="http")

    @requests_mock.Mocker()
    def test_list(self, m):
        m.get(
            "http://localhost/api/plugins?embedded=0",
            json=[
                {
                    "name": "AWS Data Sources",
                    "type": "app",
                    "id": "aws-datasource-provisioner-app",
                    "enabled": "true",
                    "pinned": "true",
                    "info": {
                        "author": {"name": "Grafana Labs", "url": "https://grafana.com/"},
                        "description": "AWS Datasource Provisioner",
                        "links": [],
                        "logos": {
                            "small": "public/plugins/aws-datasource-provisioner-app/assets/logo-small.svg",
                            "large": "public/plugins/aws-datasource-provisioner-app/assets/logo-large.svg",
                        },
                        "build": {
                            "time": 1684358005873,
                            "repo": "https://github.com/grafana/aws-datasource-provisioner-app",
                            "branch": "main",
                            "hash": "cae4b5ad22e487e9e0f81fca8ed801b36f87c733",
                        },
                        "screenshots": "",
                        "version": "1.8.1",
                        "updated": "2023-05-17",
                    },
                    "dependencies": {"grafanaDependency": ">=7.3.0", "grafanaVersion": ">=7.3.0", "plugins": []},
                    "latestVersion": "",
                    "hasUpdate": "false",
                    "defaultNavUrl": "/plugins/aws-datasource-provisioner-app/page/aws-services",
                    "category": "",
                    "state": "",
                    "signature": "valid",
                    "signatureType": "grafana",
                    "signatureOrg": "Grafana Labs",
                },
                {
                    "name": "AWS IoT SiteWise",
                    "type": "datasource",
                    "id": "grafana-iot-sitewise-datasource",
                    "enabled": "true",
                    "pinned": "false",
                    "info": {
                        "author": {"name": "Grafana Labs", "url": "https://grafana.com"},
                        "description": "A managed service to collect, store, organize and monitor data from industrial "
                        "equipment",
                        "links": [
                            {"name": "Website", "url": "https://aws.amazon.com/iot-sitewise/"},
                            {
                                "name": "Issue Tracker",
                                "url": "https://github.com/grafana/iot-sitewise-datasource/issues",
                            },
                        ],
                        "logos": {
                            "small": "public/plugins/grafana-iot-sitewise-datasource/img/sitewise.svg",
                            "large": "public/plugins/grafana-iot-sitewise-datasource/img/sitewise.svg",
                        },
                        "build": {
                            "time": 1692806526291,
                            "repo": "https://github.com/grafana/iot-sitewise-datasource",
                            "branch": "main",
                            "hash": "a56fdf241cf677e025530d484472a0c838f05b79",
                        },
                        "screenshots": [],
                        "version": "1.11.0",
                        "updated": "2023-08-23",
                    },
                    "dependencies": {"grafanaDependency": ">=8.5.0", "grafanaVersion": "8.5.0", "plugins": []},
                    "latestVersion": "",
                    "hasUpdate": "false",
                    "defaultNavUrl": "/plugins/grafana-iot-sitewise-datasource/",
                    "category": "iot",
                    "state": "",
                    "signature": "unsigned",
                    "signatureType": "",
                    "signatureOrg": "",
                },
                {
                    "name": "AWS IoT TwinMaker App",
                    "type": "app",
                    "id": "grafana-iot-twinmaker-app",
                    "enabled": "true",
                    "pinned": "true",
                    "info": {
                        "author": {"name": "AWS IoT TwinMaker", "url": "https://aws.amazon.com"},
                        "description": "Create end-user 3D digital twin applications to monitor industrial operations "
                        "with AWS IoT TwinMaker. AWS IoT TwinMaker is a service that makes it faster and"
                        "easier for developers to create digital replicas of real-world systems, "
                        "helping more customers realize the potential of digital twins to optimize operations.",
                        "links": [{"name": "Website", "url": "https://github.com/grafana/grafana-iot-twinmaker-app"}],
                        "logos": {
                            "small": "public/plugins/grafana-iot-twinmaker-app/img/AWS-IoT-TwinMaker.png",
                            "large": "public/plugins/grafana-iot-twinmaker-app/img/AWS-IoT-TwinMaker.png",
                        },
                        "build": {
                            "time": 1684967559156,
                            "repo": "https://github.com/grafana/grafana-iot-twinmaker-app",
                            "branch": "main",
                            "hash": "bc9c49f1b66c6146d17667e733b319eae967504b",
                        },
                        "screenshots": [],
                        "version": "1.6.2",
                        "updated": "2023-05-24",
                    },
                    "dependencies": {"grafanaDependency": ">=8.4.0", "grafanaVersion": "8.4.0", "plugins": []},
                    "latestVersion": "",
                    "hasUpdate": "false",
                    "defaultNavUrl": "/plugins/grafana-iot-twinmaker-app/",
                    "category": "iot",
                    "state": "",
                    "signature": "unsigned",
                    "signatureType": "",
                    "signatureOrg": "",
                },
                {
                    "name": "Akumuli",
                    "type": "datasource",
                    "id": "akumuli-datasource",
                    "enabled": "true",
                    "pinned": "false",
                    "info": {
                        "author": {"name": "Eugene Lazin", "url": "https://akumuli.org"},
                        "description": "Datasource plugin for Akumuli time-series database",
                        "links": [{"name": "Project site", "url": "https://github.com/akumuli/Akumuli"}],
                        "logos": {
                            "small": "public/plugins/akumuli-datasource/img/logo.svg.png",
                            "large": "public/plugins/akumuli-datasource/img/logo.svg.png",
                        },
                        "build": {},
                        "screenshots": "",
                        "version": "1.3.12",
                        "updated": "2019-12-19",
                    },
                    "dependencies": {"grafanaDependency": "", "grafanaVersion": "4.5.x", "plugins": []},
                    "latestVersion": "",
                    "hasUpdate": "false",
                    "defaultNavUrl": "/plugins/akumuli-datasource/",
                    "category": "",
                    "state": "",
                    "signature": "unsigned",
                    "signatureType": "",
                    "signatureOrg": "",
                },
                {
                    "name": "Alert list",
                    "type": "panel",
                    "id": "alertlist",
                    "enabled": "true",
                    "pinned": "false",
                    "info": {
                        "author": {"name": "Grafana Labs", "url": "https://grafana.com"},
                        "description": "Shows list of alerts and their current status",
                        "links": "null",
                        "logos": {
                            "small": "public/app/plugins/panel/alertlist/img/icn-singlestat-panel.svg",
                            "large": "public/app/plugins/panel/alertlist/img/icn-singlestat-panel.svg",
                        },
                        "build": {},
                        "screenshots": "null",
                        "version": "",
                        "updated": "",
                    },
                    "dependencies": {"grafanaDependency": "", "grafanaVersion": "*", "plugins": []},
                    "latestVersion": "",
                    "hasUpdate": "false",
                    "defaultNavUrl": "/plugins/alertlist/",
                    "category": "",
                    "state": "",
                    "signature": "internal",
                    "signatureType": "",
                    "signatureOrg": "",
                },
            ],
        )
        plugins = self.grafana.plugin.list()
        self.assertTrue(len(plugins), 6)

    @requests_mock.Mocker()
    def test_install(self, m):
        m.post("http://localhost/api/plugins/alertlist/install", json={"message": "Plugin alertlist installed"})
        response = self.grafana.plugin.install(plugin_id="alertlist", version="1.3.12")
        self.assertEqual(response["message"], "Plugin alertlist installed")

    @requests_mock.Mocker()
    def test_uninstall(self, m):
        m.post("http://localhost/api/plugins/alertlist/uninstall", json={"message": "Plugin alertlist uninstalled"})
        response = self.grafana.plugin.uninstall(plugin_id="alertlist")
        self.assertEqual(response["message"], "Plugin alertlist uninstalled")

    @requests_mock.Mocker()
    def test_health(self, m):
        m.get("http://localhost/api/plugins/alertlist/health", json={"message": "Plugin alertlist healthy"})
        response = self.grafana.plugin.health(plugin_id="alertlist")
        self.assertEqual(response["message"], "Plugin alertlist healthy")

    @requests_mock.Mocker()
    def test_metrics(self, m):
        m.get("http://localhost/api/plugins/grafana-timestream-datasource/metrics", json={"message": "Not found"})
        response = self.grafana.plugin.metrics(plugin_id="grafana-timestream-datasource")
        self.assertEqual(response["message"], "Not found")



import unittest

import requests_mock

from grafana_client import GrafanaApi
from grafana_client.model import PersonalPreferences


class UsersTestCase(unittest.TestCase):
    def setUp(self):
        self.grafana = GrafanaApi(("admin", "admin"), host="localhost", url_path_prefix="", protocol="http")

    @requests_mock.Mocker()
    def test_update_user(self, m):
        m.put(
            "http://localhost/api/users/foo",
            json={},
        )
        user = self.grafana.users.update_user("foo", {})
        self.assertEqual(user, {})

    @requests_mock.Mocker()
    def test_get_user(self, m):
        m.get(
            "http://localhost/api/users/foo",
            json={},
        )
        user = self.grafana.users.get_user("foo")
        self.assertEqual(user, {})

    @requests_mock.Mocker()
    def test_find_user(self, m):
        m.get(
            "http://localhost/api/users/lookup?loginOrEmail=foo",
            json={},
        )
        user = self.grafana.users.find_user("foo")
        self.assertEqual(user, {})

    @requests_mock.Mocker()
    def test_search_users_default(self, m):
        m.get(
            "http://localhost/api/users?query=foo&page=1",
            json=[{"name": "foo"}, {"name": "bar"}],
        )
        m.get(
            "http://localhost/api/users?query=foo&page=2",
            json=[],
        )
        users = self.grafana.users.search_users("foo")
        self.assertEqual(len(users), 2)

    @requests_mock.Mocker()
    def test_search_users_default_empty(self, m):
        m.get(
            "http://localhost/api/users?query=foo&page=1",
            json=[],
        )
        users = self.grafana.users.search_users("foo")
        self.assertEqual(len(users), 0)

    @requests_mock.Mocker()
    def test_search_users_page2(self, m):
        m.get(
            "http://localhost/api/users?query=foo&page=2",
            json=[{}],
        )
        users = self.grafana.users.search_users("foo", page=2)
        self.assertEqual(users, [{}])

    @requests_mock.Mocker()
    def test_search_users_perpage(self, m):
        m.get(
            "http://localhost/api/users?query=foo&page=1&perpage=1",
            json=[{"name": "foo"}, {"name": "bar"}],
        )
        m.get(
            "http://localhost/api/users?query=foo&page=2&perpage=1",
            json=[],
        )
        users = self.grafana.users.search_users("foo", perpage=1)
        self.assertEqual(len(users), 2)

    @requests_mock.Mocker()
    def test_search_users_perpage_no_endless_loop(self, m):
        m.get(
            "http://localhost/api/users?query=foo&page=1&perpage=5",
            json=[{"name": "foo"}, {"name": "bar"}],
        )
        m.get(
            "http://localhost/api/users?query=foo&page=2&perpage=5",
            json=[],
        )
        users = self.grafana.users.search_users("foo", perpage=5)
        self.assertEqual(len(users), 2)

    @requests_mock.Mocker()
    def test_get_user_organisations(self, m):
        m.get(
            "http://localhost/api/users/foo/orgs",
            json=[],
        )
        users = self.grafana.users.get_user_organisations("foo")
        self.assertEqual(users, [])


class UserTestCase(unittest.TestCase):
    def setUp(self):
        self.grafana = GrafanaApi(("admin", "admin"), host="localhost", url_path_prefix="", protocol="http")

    @requests_mock.Mocker()
    def test_get_actual_user(self, m):
        m.get(
            "http://localhost/api/user",
            json={},
        )
        result = self.grafana.user.get_actual_user()
        self.assertEqual(result, {})

    @requests_mock.Mocker()
    def test_change_actual_user_password(self, m):
        m.put(
            "http://localhost/api/user/password",
            json={},
        )
        result = self.grafana.user.change_actual_user_password("old", "new")
        self.assertEqual(result, {})

    @requests_mock.Mocker()
    def test_switch_user_organisation(self, m):
        m.post(
            "http://localhost/api/users/foo/using/acme",
            json={},
        )
        result = self.grafana.user.switch_user_organisation("foo", "acme")
        self.assertEqual(result, {})

    @requests_mock.Mocker()
    def test_switch_actual_user_organisation(self, m):
        m.post(
            "http://localhost/api/user/using/acme",
            json={},
        )
        result = self.grafana.user.switch_actual_user_organisation("acme")
        self.assertEqual(result, {})

    @requests_mock.Mocker()
    def test_get_actual_user_organisations(self, m):
        m.get(
            "http://localhost/api/user/orgs",
            json=[],
        )
        result = self.grafana.user.get_actual_user_organisations()
        self.assertEqual(result, [])

    @requests_mock.Mocker()
    def test_star_actual_user_dashboard(self, m):
        m.post(
            "http://localhost/api/user/stars/dashboard/987vb7t33",
            json={},
        )
        result = self.grafana.user.star_actual_user_dashboard("987vb7t33")
        self.assertEqual(result, {})

    @requests_mock.Mocker()
    def test_unstar_actual_user_dashboard(self, m):
        m.delete(
            "http://localhost/api/user/stars/dashboard/987vb7t33",
            json={},
        )
        result = self.grafana.user.unstar_actual_user_dashboard("987vb7t33")
        self.assertEqual(result, {})

    @requests_mock.Mocker()
    def test_get_preferences(self, m):
        m.get("http://localhost/api/user/preferences", json={"theme": "", "homeDashboardId": 0, "timezone": ""})

        result = self.grafana.user.get_preferences()
        self.assertEqual(result["homeDashboardId"], 0)

    @requests_mock.Mocker()
    def test_update_preferences(self, m):
        m.put("http://localhost/api/user/preferences", json={"message": "Preferences updated"})
        preference = self.grafana.user.update_preferences(
            PersonalPreferences(theme="", homeDashboardId=999, timezone="utc")
        )
        self.assertEqual(preference["message"], "Preferences updated")

    @requests_mock.Mocker()
    def test_patch_preferences(self, m):
        m.patch("http://localhost/api/user/preferences", json={"message": "Preferences updated"})
        preference = self.grafana.user.patch_preferences(PersonalPreferences(homeDashboardUID="zgjG8dKVz"))
        self.assertEqual(preference["message"], "Preferences updated")



import unittest

import requests_mock

from grafana_client import GrafanaApi


class DashboardVersionsTestCase(unittest.TestCase):
    def setUp(self):
        self.grafana = GrafanaApi(("admin", "admin"), host="localhost", url_path_prefix="", protocol="http")

    def test_api_path_success(self):
        api_path = self.grafana.dashboard_versions.api_path(dashboard_id=42)
        self.assertEqual(api_path, "/dashboards/id/42")

    def test_api_path_failure(self):
        with self.assertRaises(LookupError) as ctx:
            self.grafana.dashboard_versions.api_path()
        self.assertEqual(str(ctx.exception), "Either dashboard_id or dashboard_uid is required")

    @requests_mock.Mocker()
    def test_get_dashboard_versions_by_id(self, m):
        m.get(
            "http://localhost/api/dashboards/id/1/versions",
            json=[
                {
                    "id": 2,
                    "dashboardId": 1,
                    "parentVersion": 1,
                    "restoredFrom": 0,
                    "version": 2,
                    "created": "2017-06-08T17:24:33-04:00",
                    "createdBy": "admin",
                    "message": "Updated panel title",
                },
                {
                    "id": 1,
                    "dashboardId": 1,
                    "parentVersion": 0,
                    "restoredFrom": 0,
                    "version": 1,
                    "created": "2017-06-08T17:23:33-04:00",
                    "createdBy": "admin",
                    "message": "Initial save",
                },
            ],
        )
        versions = self.grafana.dashboard_versions.get_dashboard_versions_by_id(dashboard_id=1, limit=10, start=0)
        self.assertEqual(versions[0]["dashboardId"], 1)

    @requests_mock.Mocker()
    def test_get_dashboard_versions_by_uid(self, m):
        m.get(
            "http://localhost/api/dashboards/uid/QA7wKklGz/versions",
            json=[
                {
                    "id": 2,
                    "dashboardId": 1,
                    "uid": "QA7wKklGz",
                    "parentVersion": 1,
                    "restoredFrom": 0,
                    "version": 2,
                    "created": "2017-06-08T17:24:33-04:00",
                    "createdBy": "admin",
                    "message": "Updated panel title",
                },
                {
                    "id": 1,
                    "dashboardId": 1,
                    "uid": "QA7wKklGz",
                    "parentVersion": 0,
                    "restoredFrom": 0,
                    "version": 1,
                    "created": "2017-06-08T17:23:33-04:00",
                    "createdBy": "admin",
                    "message": "Initial save",
                },
            ],
        )
        versions = self.grafana.dashboard_versions.get_dashboard_versions_by_uid(
            dashboard_uid="QA7wKklGz", limit=10, start=0
        )
        self.assertEqual(versions[0]["uid"], "QA7wKklGz")

    @requests_mock.Mocker()
    def test_get_dashboard_version_by_id(self, m):
        m.get(
            "http://localhost/api/dashboards/id/1/versions/1",
            json={
                "id": 1,
                "dashboardId": 1,
                "parentVersion": 0,
                "restoredFrom": 0,
                "version": 1,
                "created": "2017-04-26T17:18:38-04:00",
                "message": "Initial save",
                "data": {"rows": [], "schemaVersion": 14, "timezone": "browser", "title": "test", "version": 1},
                "createdBy": "admin",
            },
        )
        dashboard = self.grafana.dashboard_versions.get_dashboard_version_by_id(dashboard_id=1, version_id=1)
        self.assertEqual(dashboard["dashboardId"], 1)

    @requests_mock.Mocker()
    def test_get_dashboard_version_by_uid_success(self, m):
        m.get(
            "http://localhost/api/dashboards/uid/QA7wKklGz/versions/1",
            json={
                "id": 1,
                "dashboardId": 1,
                "uid": "QA7wKklGz",
                "parentVersion": 0,
                "restoredFrom": 0,
                "version": 1,
                "created": "2017-04-26T17:18:38-04:00",
                "message": "Initial save",
                "data": {"rows": [], "schemaVersion": 14, "timezone": "browser", "title": "test", "version": 1},
                "createdBy": "admin",
            },
        )
        dashboard = self.grafana.dashboard_versions.get_dashboard_version_by_uid(
            dashboard_uid="QA7wKklGz", version_id=1
        )
        self.assertEqual(dashboard["uid"], "QA7wKklGz")

    def test_get_dashboard_version_by_uid_failure(self):
        with self.assertRaises(LookupError) as ctx:
            self.grafana.dashboard_versions.get_dashboard_version_by_uid(dashboard_uid="QA7wKklGz")
        self.assertEqual(str(ctx.exception), "version_id is required")

    @requests_mock.Mocker()
    def test_restore_dashboard_by_id(self, m):
        m.post(
            "http://localhost/api/dashboards/id/1/restore",
            json={"slug": "my-dashboard", "status": "success", "version": 3},
        )
        result = self.grafana.dashboard_versions.restore_dashboard_by_id(dashboard_id=1, version_id=1)
        self.assertEqual(result["status"], "success")

    @requests_mock.Mocker()
    def test_restore_dashboard_by_uid_success(self, m):
        m.post(
            "http://localhost/api/dashboards/uid/QA7wKklGz/restore",
            json={
                "id": 70,
                "slug": "my-dashboard",
                "status": "success",
                "uid": "QA7wKklGz",
                "url": "/d/QA7wKklGz/my-dashboard",
                "version": 3,
            },
        )
        result = self.grafana.dashboard_versions.restore_dashboard_by_uid(dashboard_uid="QA7wKklGz", version_id=1)
        self.assertEqual(result["status"], "success")
        self.assertEqual(result["uid"], "QA7wKklGz")

    def test_restore_dashboard_by_uid_failure(self):
        with self.assertRaises(LookupError) as ctx:
            self.grafana.dashboard_versions.restore_dashboard_by_uid(dashboard_uid="QA7wKklGz")
        self.assertEqual(str(ctx.exception), "version_id is required")

    @requests_mock.Mocker()
    def test_calculate_diff_success(self, m):
        m.post(
            "http://localhost/api/dashboards/calculate-diff",
            headers={"Content-Type": "text/html; charset=UTF-8"},
            text="""
<p id="l1" class="diff-line diff-json-same">
  <!-- Diff omitted -->
</p>
""",
        )
        result = self.grafana.dashboard_versions.calculate_diff(
            base_dashboard_id=1, base_version_id=1, new_dashboard_id=1, new_version_id=2
        )
        self.assertIn("diff-json", result)

    def test_calculate_diff_failure(self):
        with self.assertRaises(LookupError) as ctx:
            self.grafana.dashboard_versions.calculate_diff(
                base_dashboard_id=1, base_version_id=1, new_dashboard_id=1, new_version_id=2, diff_type="foobar"
            )
        self.assertEqual(str(ctx.exception), "diff_type must be either 'json' or 'basic'")



from .base import Base


class AlertingProvisioning(Base):
    def __init__(self, client):
        super(AlertingProvisioning, self).__init__(client)
        self.client = client

    def get_alertrules_all(self):
        """
        Gets all alert rules
        @return:
        """
        get_alertrules_all_path = "/v1/provisioning/alert-rules"
        r = self.client.GET(get_alertrules_all_path)
        return r

    def get_alertrule(self, alertrule_uid):
        """

        :param alertrule_uid:
        :return:
        """
        get_alertrule_path = "/v1/provisioning/alert-rules/%s" % alertrule_uid
        r = self.client.GET(get_alertrule_path)
        return r

    def create_alertrule(self, alertrule, disable_provenance=False):
        """
        :param alertrule:
        :param disable_provenance:
        :return:
        """
        create_alertrule_path = "/v1/provisioning/alert-rules"
        headers = {}
        if disable_provenance:
            headers["X-Disable-Provenance"] = "true"
        r = self.client.POST(create_alertrule_path, json=alertrule, headers=headers)
        return r

    def update_alertrule(self, alertrule_uid, alertrule, disable_provenance=False):
        """
        :param alertrule_uid:
        :param alertrule:
        :param disable_provenance:
        :return:
        """
        update_alertrule_path = "/v1/provisioning/alert-rules/%s" % alertrule_uid
        headers = {}
        if disable_provenance:
            headers["X-Disable-Provenance"] = "true"
        r = self.client.PUT(update_alertrule_path, json=alertrule, headers=headers)
        return r

    def get_rule_group(self, folder_uid, group_uid):
        """
        :param folder_uid:
        :param group_uid:
        :param disable_provenance:
        :return:
        """
        get_rule_group_path = "/v1/provisioning/folder/%s/rule-groups/%s" % (folder_uid, group_uid)
        r = self.client.GET(get_rule_group_path)
        return r

    def update_rule_group(self, folder_uid, group_uid, alertrule_group, disable_provenance=False):
        """
        :param folder_uid:
        :param group_uid:
        :param alertrule_group:
        :return:
        """
        headers = {}
        if disable_provenance:
            headers["X-Disable-Provenance"] = "true"
        update_rule_group_path = "/v1/provisioning/folder/%s/rule-groups/%s" % (folder_uid, group_uid)
        r = self.client.PUT(update_rule_group_path, json=alertrule_group, headers=headers)
        return r

    def update_rule_group_interval(self, folder_uid, group_uid, alertrule_group, disable_provenance=False):
        """
        :param folder_uid:
        :param group_uid:
        :param alertrule_group:
        :return:
        """
        return self.update_rule_group(folder_uid, group_uid, alertrule_group, disable_provenance)

    def delete_alertrule(self, alertrule_uid):
        """
        @param alertrule_uid:
        @param alertrule:
        @return:
        """

        delete_alertrule_path = "/v1/provisioning/alert-rules/%s" % alertrule_uid
        r = self.client.DELETE(delete_alertrule_path)
        return r

    def get_contactpoints(self):
        """
        Gets all contact points
        @return:
        """
        get_contactpoints_path = "/v1/provisioning/contact-points"
        return self.client.GET(get_contactpoints_path)

    def create_contactpoint(self, contactpoint, disable_provenance=False):
        """
        Creates single contact point
        @param contactpoint:
        @param disable_provenance:
        @return:
        """
        headers = {}
        if disable_provenance:
            headers["X-Disable-Provenance"] = "true"
        create_contactpoint_path = "/v1/provisioning/contact-points"
        return self.client.POST(create_contactpoint_path, json=contactpoint, headers=headers)

    def update_contactpoint(self, contactpoint_uid, contactpoint):
        """
        Updates existing contact point
        @param contactpoint_uid:
        @param contactpoint:
        @return:
        """
        update_contactpoint_path = "/v1/provisioning/contact-points/%s" % contactpoint_uid
        return self.client.PUT(update_contactpoint_path, json=contactpoint)

    def delete_contactpoint(self, contactpoint_uid):
        """
        Deletes existing contactpoint
        @param contactpoint_uid:
        @return:
        """
        delete_contactpoint_path = "/v1/provisioning/contact-points/%s" % contactpoint_uid
        return self.client.DELETE(delete_contactpoint_path)

    def get_notification_policy_tree(self):
        """
        Gets notification policy tree
        @return:
        """
        get_notification_policy_tree_path = "/v1/provisioning/policies"
        return self.client.GET(get_notification_policy_tree_path)

    def set_notification_policy_tree(self, notification_policy_tree, disable_provenance=False):
        """
        Sets notification policy tree
        @param notification_policy_tree:
        @param disable_provenance:
        @return:
        """
        headers = {}
        if disable_provenance:
            headers["X-Disable-Provenance"] = "true"
        set_notification_policy_tree_path = "/v1/provisioning/policies"
        return self.client.PUT(set_notification_policy_tree_path, json=notification_policy_tree, headers=headers)

    def get_mute_timings(self):
        """
        Gets all mute timings
        @return:
        """
        get_mute_timings_path = "/v1/provisioning/mute-timings"
        return self.client.GET(get_mute_timings_path)

    def get_mute_timing(self, mutetiming_name):
        """
        Gets single mute timing
        @return:
        """
        get_mute_timing_path = "/v1/provisioning/mute-timings/%s" % mutetiming_name
        return self.client.GET(get_mute_timing_path)

    def create_mute_timing(self, mutetiming, disable_provenance=False):
        """
        Creates single mute timing
        @param mutetiming:
        @param disable_provenance:
        @return:
        """
        headers = {}
        if disable_provenance:
            headers["X-Disable-Provenance"] = "true"
        create_mute_timing_path = "/v1/provisioning/mute-timings"
        return self.client.POST(create_mute_timing_path, json=mutetiming, headers=headers)

    def update_mute_timing(self, mutetiming_name, mutetiming):
        """
        Updates existing mute timing
        @return:
        """
        update_mute_timing_path = "/v1/provisioning/mute-timings/%s" % mutetiming_name
        return self.client.PUT(update_mute_timing_path, json=mutetiming)

    def delete_mute_timing(self, mutetiming_name):
        """
        Deletes single mute timing
        @return:
        """
        delete_mute_timing_path = "/v1/provisioning/mute-timings/%s" % mutetiming_name
        return self.client.GET(delete_mute_timing_path)

    def get_templates(self):
        """
        Gets all templates
        @return:
        """
        get_templates_path = "/v1/provisioning/templates"
        return self.client.GET(get_templates_path)

    def get_template(self, template_name):
        """
        Gets single template
        @param template_name:
        @return:
        """
        get_template_path = "/v1/provisioning/templates/%s" % template_name
        return self.client.GET(get_template_path)

    def create_or_update_template(self, template_name, template, disable_provenance=False):
        """
        Creates or updates (if given template_name exists) template
        @param template_name:
        @param template:
        @param disable_provenance:
        @return:
        """
        headers = {}
        if disable_provenance:
            headers["X-Disable-Provenance"] = "true"
        create_template_path = "/v1/provisioning/templates/%s" % template_name
        return self.client.PUT(create_template_path, json=template, headers=headers)

    def delete_template(self, template_name):
        """
        Deletes template
        @param template_name:
        @param template:
        @return:
        """
        create_template_path = "/v1/provisioning/templates/%s" % template_name
        return self.client.DELETE(create_template_path)



"""
About
=====
Knowledgebase module for filling the missing gaps to access the Grafana API
more efficiently.
"""
from datetime import datetime
from typing import Dict, Optional, Union

from grafana_client.model import DatasourceModel


def datasource_factory(datasource: DatasourceModel) -> DatasourceModel:
    """
    Create payload suitable for creating a Grafana data source item.

    Some data sources need additional configuration beyond the bare minimum
    attributes required by `DatasourceModel`.

    TODO: This is by far not a generic or complete implementation.
          It merely satisfies the use case where Docker containers are started
          on localhost, like the data source health check demo program does.
          Many attributes are hardcoded to specific values. Additional
          parameterization work might make this function more generic in the
          future.

    TODO: Complete the list for all popular databases.
    """
    if datasource.type == "__NEVER__":  # pragma:nocover
        raise NotImplementedError("__NEVER__")

    elif datasource.type == "cratedb":
        datasource.type = "postgres"
        datasource.user = "crate"
        datasource.jsonData = {
            "postgresVersion": 1200,
            "sslmode": "disable",
        }
    elif datasource.type == "elasticsearch":
        datasource.access = "proxy"
        datasource.database = "testdrive"
        datasource.jsonData = {
            "esVersion": "7.10.0",
            "timeField": "@timestamp",
        }
    elif datasource.type == "graphite":
        datasource.access = "proxy"
        datasource.jsonData = {
            "graphiteVersion": "1.1",
        }
    elif datasource.type in ["influxdb", "influxdb+influxql"]:
        datasource.type = "influxdb"
        datasource.access = "proxy"
        datasource.jsonData = {
            "httpMode": "POST",
            "version": "InfluxQL",
        }
    elif datasource.type in ["influxdb+flux"]:
        datasource.type = "influxdb"
        datasource.access = "proxy"
        datasource.jsonData = {
            "httpMode": "POST",
            "organization": "example",
            "version": "Flux",
        }
        datasource.secureJsonData = {
            "token": "admintoken",
        }
        datasource.secureJsonFields = {
            "token": False,
        }
    elif datasource.type == "jaeger":
        datasource.access = "proxy"
    elif datasource.type == "opentsdb":
        datasource.access = "proxy"
        datasource.jsonData = {
            "tsdbVersion": 3,
        }
    elif datasource.type == "loki":
        datasource.access = "proxy"
    elif datasource.type == "mssql":
        datasource.access = "proxy"
        datasource.database = "testdrive"
        datasource.user = "sa"
        datasource.jsonData = {
            "authenticationType": "SQL Server Authentication",
        }
        datasource.secureJsonData = {
            "password": "root123?",
        }
        datasource.secureJsonFields = {
            "password": False,
        }
    elif datasource.type == "mysql":
        datasource.user = "root"
        datasource.secureJsonData = {
            "password": "root",
        }
    elif datasource.type == "postgres":
        datasource.user = "postgres"
        datasource.jsonData = {
            "postgresVersion": 1200,
            "sslmode": "disable",
        }
    elif datasource.type == "prometheus":
        datasource.access = "proxy"
    elif datasource.type == "tempo":
        datasource.access = "proxy"
    elif datasource.type == "testdata":
        pass
    elif datasource.type == "zipkin":
        datasource.access = "proxy"
    else:
        raise NotImplementedError(f"Unknown data source type: {datasource.type}")
    return datasource


def query_factory(datasource, model: Optional[dict] = None, expression: Optional[str] = None) -> Union[Dict, str]:
    """
    Create payload suitable for running a query against a Grafana data source.

    TODO: This is by far not complete. It has to be made more elaborate in order
          to query different data source types.

    TODO: Complete the list for all popular databases.
    """

    model = model or {}
    if "query" not in model and expression:
        model["query"] = expression

    request = {
        "method": "POST",
        "data": None,
        "params": None,
    }
    attrs = None

    datasource_type = datasource["type"]
    expression = model.get("query")
    if expression is None:
        raise KeyError("query not set")
    if datasource_type == "__NEVER__":  # pragma:nocover
        raise NotImplementedError("__NEVER__")
    elif datasource_type == "elasticsearch":
        query = expression
    elif datasource_type == "fetzerch-sunandmoon-datasource":
        query = expression
    elif datasource_type == "grafana-simple-json-datasource":
        query = expression

    elif datasource_type == "graphite":
        query = {"target": expression, "from": "-5m", "until": "now", "format": "json", "maxDataPoints": 300}
        if "time_from" in model:
            query["from"] = model["time_from"]
        if "time_to" in model:
            query["until"] = model["time_to"]

        request["data"] = query

    elif datasource_type == "influxdb":
        dialect = datasource["jsonData"].get("version", "InfluxQL")
        query = {
            "refId": "test",
            # "datasource": {
            #     "type": datasource["type"],
            #     "uid": datasource.get("uid"),
            # },
            # "datasourceId": datasource.get("id"),
        }
        if dialect == "InfluxQL":
            query.update(
                {
                    "q": expression,
                }
            )
            # this drive the how timestamp are rendered in result (string by default, or in milliseconds ms)
            request["params"] = {"epoch": "ms"}
            if "database" in datasource:
                request["params"].update({"db": datasource["database"]})
            request["data"] = query

        elif dialect == "Flux":
            query = {
                "datasource": {
                    "type": datasource["type"],
                    "uid": datasource.get("uid"),
                },
                "datasourceId": datasource.get("id"),
                # "exemplar": False,
                "query": expression,
            }

            attrs = [
                {
                    "name": "intervalMs",
                    "default": 30000,
                },
                {
                    "name": "maxDataPoints",
                    "default": 1441,
                },
                {
                    "name": "refId",
                    "default": "test",
                },
            ]
        else:
            raise KeyError(f"InfluxDB dialect '{dialect}' unknown")

    elif datasource_type == "jaeger":
        query = {}

    elif datasource_type == "loki":
        query = {
            "datasource": {
                "type": datasource["type"],
                "uid": datasource.get("uid"),
            },
            "datasourceId": datasource.get("id"),
            # "exemplar": False,
            "expr": expression,
        }

        attrs = [
            {
                "name": "intervalMs",
                "default": 60000,
            },
            {
                "name": "legendFormat",
                "default": "",
            },
            {
                "name": "maxLines",
                "default": 1000,
            },
            {
                "name": "maxDataPoints",
                "default": 1441,
            },
            {
                "name": "queryType",
                "default": "range",
            },
            {
                "name": "refId",
                "default": "test",
            },
            {
                "name": "resolution",
                "default": 1,
            },
        ]

    elif datasource_type == "opentsdb":
        query = {}

    elif datasource_type in ("postgres", "mssql", "mysql"):
        query = {
            "datasource": {
                "type": datasource["type"],
                "uid": datasource.get("uid"),
            },
            "datasourceId": datasource.get("id"),
            "rawSql": expression,
        }
        attrs = [
            {
                "name": "format",
                "default": "time_series",
                "choices": ["time_series", "table"],
                # "version": "8.0.0"
            },
            {
                "name": "intervalMs",
                "default": 30000,
            },
            {
                "name": "maxDataPoints",
                "default": 1441,
            },
            {
                "name": "refId",
                "default": None,
            },
        ]

    elif datasource_type == "prometheus":
        query = {
            "datasource": {
                "type": datasource["type"],
                "uid": datasource.get("uid"),
            },
            "datasourceId": datasource.get("id"),
            # "exemplar": False,
            "expr": expression,
        }

        attrs = [
            {
                "name": "format",
                "default": "time_series",
                "choices": ["time_series", "table", "heatmap"],
                # "version": "8.0.0"
            },
            {
                "name": "instant",
                "default": False,
            },
            {
                "name": "interval",
                "default": "",
            },
            {
                "name": "intervalFactor",
                "default": None,
            },
            {
                "name": "intervalMs",
                "default": 30000,
            },
            {
                "name": "legendFormat",
                "default": "",
            },
            {
                "name": "maxDataPoints",
                "default": None,
            },
            {
                "name": "queryType",
                "default": "timeSeriesQuery",
            },
            {
                "name": "refId",
                "default": "test",
            },
            {
                "name": "requestId",
                "default": "0test",
            },
            {
                "name": "step",
                "default": 300,
            },
            {
                "name": "utcOffsetSec",
                "default": 0,
            },
        ]

    elif datasource_type == "simpod-json-datasource":
        query = expression
    elif datasource_type == "tempo":
        query = {}
    elif datasource_type == "testdata":
        query = expression
    elif datasource_type == "zipkin":
        query = {}
    else:
        raise NotImplementedError(f"Unknown data source type: {datasource_type}")

    if attrs is not None and query is not None and isinstance(query, dict):
        for attr in attrs:
            value = attr["default"]
            if attr["name"] in model:
                tmp_value = model[attr["name"]]
                if "choices" in attr:
                    if tmp_value in attr["choices"]:
                        value = tmp_value
                else:
                    value = tmp_value
            if value is not None:
                query[attr["name"]] = value

        if "time_from" not in model or "time_to" not in model:
            now = datetime.now()
            if "time_from" not in model:
                model["time_from"] = int(now.timestamp()) - 5 * 60
            if "time_to" not in model:
                model["time_to"] = int(now.timestamp())

        if "instant" in query and query["instant"]:
            model["time_from"] = model["time_to"]

        payload = {
            "queries": [query],
            "from": str(model["time_from"] * 1000),
            "to": str(model["time_to"] * 1000),
        }
        request["data"] = payload

    return request


# Define health-check status queries for all database types.
# TODO: Complete the list for all popular databases.
HEALTHCHECK_EXPRESSION_MAP = {
    "elasticsearch": "url:///datasources/proxy/{datasource_id}/{database_name}/_mapping",
    "fetzerch-sunandmoon-datasource": "url:///datasources/uid/{datasource_uid}",
    "grafana-simple-json-datasource": "url:///datasources/proxy/{datasource_id}",
    # From play.grafana.org, to explore Graphite datasource.
    "graphite": "random-walk.count;dc=asia-1;app=collector;server=000",
    "influxdb": "SHOW RETENTION POLICIES on _internal",
    "influxdb+influxql": "SHOW RETENTION POLICIES on _internal",
    "influxdb+flux": "buckets()",
    "jaeger": "url:///datasources/proxy/{datasource_id}/api/services",
    "loki": 'count_over_time({job=~".+"} [5m])',
    "mssql": "SELECT 1",
    "mysql": "SELECT 1",
    "opentsdb": "url:///datasources/proxy/{datasource_id}/api/suggest?type=metrics&q=cpu&max=100",
    "postgres": "SELECT 1",
    "prometheus": "1+1",
    "simpod-json-datasource": "url:///datasources/proxy/{datasource_id}",
    "tempo": "url:///datasources/proxy/{datasource_id}/api/echo",
    "testdata": "url:///datasources/uid/{datasource_uid}",
    "zipkin": "url:///datasources/proxy/{datasource_id}/api/v2/services",
}


def get_healthcheck_expression(datasource_type: str, datasource_dialect: str = None) -> str:
    """
    Produce data source health check query for corresponding database type.
    """
    if datasource_type == "influxdb" and datasource_dialect == "Flux":
        datasource_type = "influxdb+flux"
    if datasource_type not in HEALTHCHECK_EXPRESSION_MAP:
        raise NotImplementedError(f"Unknown data source type: {datasource_type}")
    return HEALTHCHECK_EXPRESSION_MAP[datasource_type]



"""
Example program for probing whether a data source is healthy.

Documentation: https://github.com/panodata/grafana-client/blob/main/examples/datasource-health-probe.rst
"""
import json
import logging
import sys
from optparse import OptionParser

import requests
from verlib2 import Version

from grafana_client import GrafanaApi
from grafana_client.client import GrafanaClientError
from grafana_client.knowledge import datasource_factory
from grafana_client.model import DatasourceModel
from grafana_client.util import setup_logging

logger = logging.getLogger(__name__)


VERSION_7 = Version("7")
VERSION_8 = Version("8")
VERSION_9 = Version("9")


def ensure_datasource(grafana: GrafanaApi, datasource: DatasourceModel):
    """
    Ensure data source exists and is configured like desired.

    Either create the data source, or update it. The data source name will be
    used as data source identified.
    """

    datasource_name = datasource.name

    # Create data source.
    datasource = datasource.asdict()
    try:
        logger.info(f"Creating data source '{datasource_name}'")
        datasource = grafana.datasource.create_datasource(datasource)["datasource"]
    except GrafanaClientError as ex:
        # When data source already exists, update data source.
        if ex.status_code == 409:
            logger.info(f"Data source already exists: {ex.response}. Updating.")
            datasource_existing = grafana.datasource.get_datasource_by_name(datasource_name=datasource_name)
            datasource = grafana.datasource.update_datasource(datasource_existing["id"], datasource)["datasource"]
        else:
            logger.error(
                f"Failed to create or update data source '{datasource}'. "
                f"Reason: {ex.message}. Response: {ex.response}"
            )
            raise
    return datasource


def health_probe(grafana: GrafanaApi, datasource: DatasourceModel):
    """
    Add a data source dynamically, run a data source health check on it,
    and delete it again. Be graceful if the data source exists already.
    """
    # Create data source.
    datasource = ensure_datasource(grafana, datasource)
    datasource_uid = datasource["uid"]

    # Invoke the health check.
    health_info = grafana.datasource.health_inquiry(datasource_uid=datasource_uid)

    # Delete data source again.
    grafana.datasource.delete_datasource_by_uid(datasource_uid)

    return health_info.asdict_compact()


def prometheus_demo(grafana: GrafanaApi):
    datasource = DatasourceModel(
        name="probe-prometheus", type="prometheus", url="http://host.docker.internal:9090", access="server"
    )
    health_info = health_probe(grafana, datasource)
    return health_info


def run(grafana: GrafanaApi, grafana_version: Version = None):
    # When called without options, invoke the Prometheus demo.
    if len(sys.argv) == 1:
        if grafana_version < VERSION_8:
            raise NotImplementedError(
                f"Data source health check subsystem on Grafana version {grafana_version} not supported for Prometheus"
            )
        health_info = prometheus_demo(grafana)

    # When called with options,
    else:
        # parse them, and
        parser = OptionParser()
        parser.add_option("--type", dest="type", help="Data source type")
        parser.add_option("--url", dest="url", help="Data source URL")
        (options, args) = parser.parse_args()
        if (not options.type or not options.url) and not options.type == "testdata":
            parser.error("Options --type and --url required")

        # Sanity checks
        if options.type == "prometheus" and grafana_version < VERSION_8:
            raise NotImplementedError(
                f"Data source health check subsystem on Grafana version {grafana_version} not supported for Prometheus"
            )

        # ... create a dynamic data source with the corresponding values.
        name = f"probe-{options.type}"
        datasource = DatasourceModel(name=name, type=options.type, url=options.url, access="server")
        datasource = datasource_factory(datasource)

        # Invoke the health probe.
        health_info = health_probe(grafana, datasource)

    # Display the outcome and terminate program based on success state.
    print(json.dumps(health_info, indent=2))
    if not health_info["success"]:
        sys.exit(1)


if __name__ == "__main__":
    setup_logging(level=logging.DEBUG)

    # Connect to Grafana instance and run health probe.
    grafana_client = GrafanaApi.from_env()

    try:
        grafana_client.connect()
    except requests.exceptions.ConnectionError:
        logger.exception("Connecting to Grafana failed")
        raise SystemExit(1)

    grafana_version = Version(grafana_client.version)
    if grafana_version < VERSION_7:
        raise NotImplementedError(f"Data source health check subsystem not ready for Grafana version {grafana_version}")

    run(grafana_client, grafana_version=grafana_version)



import unittest

from parameterized import parameterized

from grafana_client.knowledge import (
    HEALTHCHECK_EXPRESSION_MAP,
    datasource_factory,
    get_healthcheck_expression,
    query_factory,
)
from grafana_client.model import DatasourceModel

SUPPORTED_DATA_SOURCE_TYPES_HEALTH = list(HEALTHCHECK_EXPRESSION_MAP.keys())
SUPPORTED_DATA_SOURCE_TYPES_DSFACTORY = SUPPORTED_DATA_SOURCE_TYPES_HEALTH.copy()
SUPPORTED_DATA_SOURCE_TYPES_DSFACTORY.append("cratedb")
SUPPORTED_DATA_SOURCE_TYPES_DSFACTORY.remove("fetzerch-sunandmoon-datasource")
SUPPORTED_DATA_SOURCE_TYPES_DSFACTORY.remove("grafana-simple-json-datasource")
SUPPORTED_DATA_SOURCE_TYPES_DSFACTORY.remove("simpod-json-datasource")


class KnowledgebaseTestCase(unittest.TestCase):
    @parameterized.expand(SUPPORTED_DATA_SOURCE_TYPES_DSFACTORY)
    def test_datasource_factory_success(self, data_source_type):
        ds = datasource_factory(DatasourceModel(name="foo", type=data_source_type, url=None, access=None))
        self.assertIsInstance(ds, DatasourceModel)

    def test_datasource_factory_unknown_failure(self):
        self.assertRaises(
            NotImplementedError,
            lambda: datasource_factory(DatasourceModel(name="foo", type="unknown", url=None, access=None)),
        )

    @parameterized.expand(SUPPORTED_DATA_SOURCE_TYPES_DSFACTORY)
    def test_query_factory_success(self, data_source_type):
        ds = datasource_factory(DatasourceModel(name="foo", type=data_source_type, url=None, access=None))
        query = query_factory(ds.asdict(), expression="bar")
        self.assertIsInstance(query, (dict, str))

    def test_query_factory_unknown_type_failure(self):
        ds = datasource_factory(DatasourceModel(name="foo", type="prometheus", url=None, access=None))
        ds.type = "unknown"
        self.assertRaises(
            NotImplementedError,
            lambda: query_factory(ds.asdict(), expression="bar"),
        )

    def test_query_factory_unknown_influxdb_dialect_failure(self):
        ds = datasource_factory(DatasourceModel(name="foo", type="influxdb", url=None, access=None))
        ds.jsonData["version"] = "unknown"
        self.assertRaises(
            KeyError,
            lambda: query_factory(ds.asdict(), expression="bar"),
        )

    @parameterized.expand(SUPPORTED_DATA_SOURCE_TYPES_HEALTH)
    def test_get_healthcheck_expression_all(self, data_source_type):
        result = get_healthcheck_expression(data_source_type)
        self.assertIsInstance(result, str)
        self.assertTrue(len(result) > 1)

    def test_get_healthcheck_expression_influxdb_flux(self):
        result = get_healthcheck_expression("influxdb", "Flux")
        self.assertEqual(result, "buckets()")

    def test_get_healthcheck_expression_unknown_failure(self):
        self.assertRaises(NotImplementedError, lambda: get_healthcheck_expression("foobar"))



import unittest

import requests_mock

from grafana_client import GrafanaApi


class RbacTestCase(unittest.TestCase):
    def setUp(self):
        self.grafana = GrafanaApi(("admin", "admin"), host="localhost", url_path_prefix="", protocol="http")

    @requests_mock.Mocker()
    def test_get_rbac_roles_all(self, m):
        m.get(
            "http://localhost/api/access-control/roles",
            json=[
                {
                    "version": 5,
                    "uid": "vi9mlLjGz",
                    "name": "fixed:datasources.permissions:writer",
                    "description": "Create, read or delete data source permissions.",
                    "global": True,
                    "updated": "2021-05-13T22:41:49+02:00",
                    "created": "2021-05-13T16:24:26+02:00",
                }
            ],
        )
        roles = self.grafana.rbac.get_rbac_roles_all()
        self.assertEqual(roles[0]["name"], "fixed:datasources.permissions:writer")
        self.assertEqual(len(roles), 1)

    @requests_mock.Mocker()
    def test_add_rbac_role_teams(self, m):
        m.post(
            "http://localhost/api/access-control/teams/1/roles",
            json={"message": "Role added to the team."},
        )
        history = m.request_history

        r = self.grafana.rbac.add_rbac_role_team("1", "XvHQJq57z")
        self.assertEqual(history[0].json()["roleUid"], "XvHQJq57z")
        self.assertEqual(r["message"], "Role added to the team.")

    @requests_mock.Mocker()
    def test_add_rbac_roles_teams(self, m):
        m.put(
            "http://localhost/api/access-control/teams/1/roles",
            json={"message": "Team roles have been updated."},
        )
        history = m.request_history

        roleUids = ["ZiHQJq5nk", "GzNQ1357k"]
        r = self.grafana.rbac.add_rbac_roles_team("1", roleUids)
        self.assertEqual(history[0].json()["roleUids"], roleUids)
        self.assertEqual(r["message"], "Team roles have been updated.")

    @requests_mock.Mocker()
    def test_remove_rbac_role_team(self, m):
        m.delete(
            "http://localhost/api/access-control/teams/1/roles/AFUXBHKnk",
            json={"message": "Role removed from team."},
        )
        r = self.grafana.rbac.remove_rbac_role_team("1", "AFUXBHKnk")
        self.assertEqual(r["message"], "Role removed from team.")



"""
Example program for checking whether a data source is healthy.

Documentation: https://github.com/panodata/grafana-client/blob/main/examples/datasource-health-check.rst
"""
import json
import logging
import sys
from optparse import OptionParser

import requests
from verlib2 import Version

from grafana_client import GrafanaApi
from grafana_client.model import DatasourceHealthResponse
from grafana_client.util import setup_logging

logger = logging.getLogger(__name__)

VERSION_7 = Version("7")


def run(grafana: GrafanaApi):
    parser = OptionParser()
    parser.add_option("--uid", dest="uid", help="Data source UID")
    (options, args) = parser.parse_args()

    if options.uid:
        datasources = [grafana.datasource.get_datasource_by_uid(options.uid)]
    else:
        datasources = grafana.datasource.list_datasources()

    if not datasources:
        logger.warning(f"No data sources found at {grafana.url}")
        sys.exit(2)

    logger.info(f"Probing {len(datasources)} data sources")

    success = True
    statistics = {"ok": 0, "error": 0, "fatal": 0, "unknown": 0}
    for datasource in datasources:
        logger.info(f"Discovered datasource with uid={datasource['uid']} and type={datasource['type']}")

        # Invoke the health check.
        try:
            health_info = grafana.datasource.health_inquiry(datasource_uid=datasource["uid"])
            if health_info.success:
                statistics["ok"] += 1
            else:
                if health_info.status == "FATAL":
                    statistics["fatal"] += 1
                else:
                    statistics["error"] += 1
        except NotImplementedError as ex:
            message = f"{ex.__class__.__name__}: {ex}"
            health_info = DatasourceHealthResponse(
                uid=datasource["uid"], type=datasource["type"], success=False, status="UNKNOWN", message=message
            )
            statistics["unknown"] += 1
            logger.exception(message)

        # Display the outcome and terminate program based on success state.
        print(json.dumps(health_info.asdict_compact(), indent=2))
        if not health_info.success:
            success = False

    logger.info(f"Statistics: {statistics}")

    if not success:
        sys.exit(1)


if __name__ == "__main__":
    setup_logging(level=logging.DEBUG)

    # Connect to Grafana instance and run health check.
    grafana_client = GrafanaApi.from_env()

    try:
        grafana_client.connect()
    except requests.exceptions.ConnectionError:
        logger.exception("Connecting to Grafana failed")
        raise SystemExit(1)

    grafana_version = Version(grafana_client.version)
    if grafana_version < VERSION_7:
        raise NotImplementedError(f"Data source health check subsystem not ready for Grafana version {grafana_version}")

    run(grafana_client)



"""
About
=====

Example program for interacting with the Grafana Health API endpoint, in order
to inquire the Grafana version.

Usage
=====

Make sure to adjust the options to `GrafanaApi.from_url(...)` at the bottom of the
file before running the program.

Synopsis
========
::

    source .venv/bin/activate
    python examples/grafana-health.py

"""
from grafana_client import GrafanaApi
from grafana_client.client import TokenAuth


def run_conversation(grafana: GrafanaApi):
    print("Grafana address")
    print(grafana.client.url)

    print("Health check")
    print(grafana.health.check())


if __name__ == "__main__":
    # Connect to custom Grafana instance.
    grafana = GrafanaApi.from_url(
        url="http://localhost:3000/",
        credential=TokenAuth(token="eyJrIjoiMGJIcDBZZVdGV3VNWHYyV1J2dU5lcnVBSW1qYzR2c1MiLCJuIjoiZm9vIiwiaWQiOjF9"),
    )

    # Connect to Grafana instance of Grafana Labs fame.
    # grafana = GrafanaApi(host="play.grafana.org", protocol="https")

    run_conversation(grafana)
